/*
 * Worldvisions Weaver Software:
 *   Copyright (C) 1998, 1999 Worldvisions Computer Technology, Inc.
 *
 * The TCP-related parts of Fast Forward.
 *
 */

#include "wvfftcp.h"

#include <time.h>


///////////////////////////////////////////////////////////
// WvFF_TCPListener stuff
///////////////////////////////////////////////////////////

WvFF_TCPListener::WvFF_TCPListener( const WvIPPortAddr& l_port,
				    WvFastForward *	forward,
				    const WvString& 	dest,
				    bool		transparent )
/*******************************************************************/
: WvTCPListener( l_port ), fwd( *forward ), dest_str( dest ),
    transproxy( transparent )
{
    setcallback( tcp_accept_callback, NULL );
}


// from wvtcp.cc -- yuck!
void nice_tcpopts(int fd);


WvFF_TCPConn * WvFF_TCPListener::accept_ff()
/******************************************/
{
    struct sockaddr_in sin;
    size_t len = sizeof( sin );
    int    new_fd;
    WvFF_TCPConn * ret;

    // FIXME!  WvFF_TCPConn should be a WvStreamClone of a WvTCPConn.
    // This kind of direct access to ::accept() will bite us sooner or
    // later...
    new_fd = ::accept( fd, (struct sockaddr *) &sin, &len );
    nice_tcpopts(fd);
    ret = new WvFF_TCPConn( new_fd, WvIPPortAddr( &sin ) );
    return( ret );
}

void WvFF_TCPListener::tcp_accept_callback( WvStream& s, void * userdata )
/************************************************************************/
{
    WvFF_TCPListener& lstnr = *(WvFF_TCPListener *) &s;

    WvFF_TCPConn * origin = lstnr.accept_ff();

    if( lstnr.transproxy )
    	// If this was a transparent proxy, we need to forward the connection
    	// to the LOCAL address of the origin... (since we're pretending to
    	// be the destination.)
    	lstnr.fwd.log( "Transparent proxy from %s to %s...\n",
    			*origin->src(), origin->localaddr() );
    else
    	// If it WASN'T a transparent proxy, it's pretty easy, just forward
    	// it to the listener's stored destination.
	lstnr.fwd.log( "Connection from %s to %s...\n",
    			*origin->src(), *lstnr.src() );


    // If there are too many connections open, close this new one!
    if( lstnr.fwd.too_many_conns() ) {
    	lstnr.fwd.err( "Not forwarding -- too many connections open!\n" );
	delete origin;
    	return;
    }

    // Must now start the outgoing connection, after we find who it should
    // go to.
    WvIPPortAddr dest_addr;

    if( lstnr.transproxy && 
    		( (WvIPPortAddr *)lstnr.src() )->port 
    	     != origin->localaddr().port )

    	dest_addr = origin->localaddr();
    else {
	char *	 err_buf;
	if( lstnr.transproxy && lstnr.dest_str == "" ) {
	    lstnr.fwd.err( "Transparent proxy would form a loop.  Dropped." );
	    delete origin;
	    return;
	}
	if( make_port_addr( dest_addr, lstnr.dest_str, &err_buf ) == false ) {
    	    lstnr.fwd.err( err_buf );
	    delete origin;
    	    return;
    	}
    }

    WvFF_TCPConn * dest = new WvFF_TCPConn( dest_addr, origin );
    if( !dest ) {
    	lstnr.fwd.err( "Connection to %s dropped!\n",
    				*(WvIPPortAddr *) lstnr.src() );
	delete origin;
	return;
    } else if( dest->isok() == false ) {
    	lstnr.fwd.err( "Can't forward to %s: %s\n", 
    				*(WvIPPortAddr *) lstnr.src(), dest->errstr() );
	delete origin;
	return;
    } else {
    	lstnr.fwd.log( "...forwarding to %s.\n", dest_addr );
    }

    // Now the connection we just made knows about the incoming one, but the
    // incoming one doesn't know about the one we just made.  Fix:
    origin->remote_conn = dest;

    // Tell Fast Forward about both new connections.
    lstnr.fwd.get_tcp().append( origin, true );
    lstnr.fwd.get_tcp().append( dest, true );
}


///////////////////////////////////////////////////////////
// WvFF_TCPConn stuff
///////////////////////////////////////////////////////////

unsigned int WvFF_TCPConn::conn_count = 0;

WvFF_TCPConn::WvFF_TCPConn( int open_fd, const WvIPPortAddr& rem )
/****************************************************************/
// protected.  Used by WvFF_TCPListener::accept_ff().
: WvTCPConn( open_fd, rem )
{
    last_xmit   = time( NULL );
    remote_conn = NULL;
    setcallback( tcp_callback, NULL );
    inc_num_conns();
}

WvFF_TCPConn::WvFF_TCPConn( const WvIPPortAddr& rem, WvFF_TCPConn * rem_conn )
/****************************************************************************/
// public.  Used above in WvFF_TCPListener::tcp_accept_callback().
: WvTCPConn( rem )
{
    last_xmit   = time( NULL );
    remote_conn = rem_conn;
    setcallback( tcp_callback, NULL );
    inc_num_conns();
}

WvFF_TCPConn::~WvFF_TCPConn()
/***************************/
{
    if (remote_conn)
    {
	remote_conn->close();
	remote_conn->remote_conn = NULL;
    }
    dec_num_conns();
}

void WvFF_TCPConn::tcp_callback( WvStream& s, void * userdata )
/*************************************************************/
// Standard stream callback function.  We receive something from s, and we
// need to pipe it through to the remote stream.
{
    WvFF_TCPConn&   conn    = *(WvFF_TCPConn *) &s;
    WvFF_TCPConn *  out     = conn.remote_conn;
    WvBuffer&       out_buf = out->write_buf;
    char	    buf[ 1024 ];
    int		    num_read;
    
    // first, try to clear out our own write buffer
    while( conn.write_buf.used() && conn.select( 0, false, true) ) {
	size_t size = conn.write_buf.used();
	conn.write( conn.write_buf.get( size ), size );
	conn.last_xmit = time( NULL );
	out->last_xmit = time( NULL );
    }

    while( out_buf.used() || conn.select( 0 ) ) {
    	if( out_buf.used() != 0 ) {
    	    if( !out->select( 0, false, true ) )
    	    	return;

    	    size_t size = out_buf.used();
    	    out->write( out_buf.get( size ), size );
    	    out->last_xmit = time( NULL );
    	    conn.last_xmit = time( NULL );
    	    continue;
    	}

    	num_read = conn.read( buf, sizeof( buf ) );

    	if( !out->select( 0, false, true ) )
    	    out_buf.put( buf, num_read );
    	else {
    	    out->write( buf, num_read );
    	    out->last_xmit = time( NULL );
    	    conn.last_xmit = time( NULL );
    	}
    }
}

